/*
 * glvideo.cpp
 *
 * Copyright (C) 2008 Christophe Thommeret <hftom@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include <QGLPixelBuffer>
#include <QGLFramebufferObject>

#include <KDebug>

#include "videowidget.h"
#include "glvideo.h"
#include "shaders.h"



GLVideo::GLVideo( VideoWidget *parent, GLFunc *glf )
{
	func = glf;
	supportedFormats = XINE_VORAW_RGB; // rgb
	vw = parent;
	PHandle = 0;
}



GLVideo::~GLVideo()
{
}



void GLVideo::initFragProg()
{
	func->mglGenPrograms( 1, &fpyv12 );
	func->mglBindProgram( GL_FRAGMENT_PROGRAM_ARB, fpyv12 );
	func->mglProgramString( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(FPyv12), FPyv12 );
	if ( glGetError()==GL_INVALID_OPERATION )
		kDebug() << glGetString( GL_PROGRAM_ERROR_STRING_ARB );

	func->mglGenPrograms( 1, &fpyuy2 );
	func->mglBindProgram( GL_FRAGMENT_PROGRAM_ARB, fpyuy2 );
	func->mglProgramString( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(FPyuy2), FPyuy2 );
	if ( glGetError()==GL_INVALID_OPERATION )
		kDebug() << glGetString( GL_PROGRAM_ERROR_STRING_ARB );
}



void GLVideo::initShader()
{
	PHandle=func->mglCreateProgramObject();
	FSHandle=func->mglCreateShaderObject( GL_FRAGMENT_SHADER_ARB );
	func->mglShaderSource( FSHandle, 1, &FSyuy2, NULL );
	func->mglCompileShader( FSHandle );
	int i;
	char *s;
	func->mglGetObjectParameteriv( FSHandle, GL_OBJECT_COMPILE_STATUS_ARB, &i );
	s=(char*)malloc( 32768 );
	func->mglGetInfoLog( FSHandle, 32768, NULL, s );
	kDebug() << "Shader Compile Log: " << s;
	free(s);
	func->mglAttachObject( PHandle, FSHandle );
	func->mglLinkProgram( PHandle );
	s=(char*)malloc( 32768 );
	func->mglGetInfoLog( PHandle, 32768, NULL, s);
	kDebug() << "Shader Link Log: " << s;
	free(s);
}



void GLVideo::initGL()
{
	func->init();

	if ( func->doVideoSync() )
		kDebug() << "can do Sync2VBlank";
	else
		kDebug() << "can NOT do Sync2VBlank";

	if ( func->doTextureRect() )
		kDebug() << "can do TextureRect";
	else
		kDebug() << "can NOT do TextureRect";

	if ( func->doFragShader() ) {
		kDebug() << "can do FragShader";
		initShader();
	}
	else
		kDebug() << "can NOT do FragShader";

	if ( func->doFragProg() ) {
		kDebug() << "can do FragProg";
		supportedFormats |= XINE_VORAW_YV12 | XINE_VORAW_YUY2;
		initFragProg();
	}
	else
		kDebug() << "can NOT do FragProg";


	if ( func->doPBO() ) {
		kDebug() << "can do PBO";
		func->mglGenBuffers( 1, &videoPBO );
	}
	else
		kDebug() << "can NOT do PBO";

	glGenTextures( 3, videoTexture );
	glGenTextures( XINE_VORAW_MAX_OVL, ovlTexture );
}



bool GLVideo::takeVideoShot( QImage &image, int width )
{
	if ( !videoFrame.width )
		return false;

	if ( QGLFramebufferObject::hasOpenGLFramebufferObjects() ) {
		kDebug() << ": framebuffer_object";
		vw->makeCurrent();
		QGLFramebufferObject *fbo = new QGLFramebufferObject( width, (int)(width/videoFrame.aspectRatio) );
		glPushAttrib( GL_ALL_ATTRIB_BITS );
		glMatrixMode( GL_PROJECTION );
		glPushMatrix();
		glViewport( 0, 0, fbo->size().width(), fbo->size().height() );
		glLoadIdentity();
		glFrustum( 0, fbo->size().width(), fbo->size().height(), 0, 4.9999, 60.0 );
		glMatrixMode( GL_MODELVIEW );
		fbo->bind();
		glLoadIdentity();
		draw( fbo->size().width(), fbo->size().height(), true, false );
		image = fbo->toImage();
		fbo->release();
		glMatrixMode( GL_PROJECTION );
		glPopMatrix();
		glPopAttrib();
		delete fbo;
		return true;
	}
	else if ( QGLPixelBuffer::hasOpenGLPbuffers() ) {
		kDebug() << ": pbuffer";
		QGLPixelBuffer *pb = new QGLPixelBuffer( width, (int)(width/videoFrame.aspectRatio), vw->format(), vw );
		pb->makeCurrent();
		glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
		glEnable( GL_TEXTURE_2D );
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
		glClear( GL_COLOR_BUFFER_BIT );
		glViewport( 0, 0, pb->size().width(), pb->size().height() );
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		glFrustum( 0, pb->size().width(), pb->size().height(), 0, 4.9999, 60.0 );
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		draw( pb->size().width(), pb->size().height(), true, false );
		image = pb->toImage();
		delete pb;
		return true;
	}
	else
		return false;
}



void GLVideo::draw( int width, int height, bool hasVideo, bool drawOvl )
{
	if ( !videoFrame.width )
		return;

	glPushMatrix();
	glTranslatef( width/2, height/2, 0 );
	glScalef( width/2, height/2, 1.0 );
	GLfloat left, right, top, bottom, u, u1, v, v1, videoWidth, videoHeight;
	GLfloat pixborderW, pixborderH;
	if ( videoFrame.texFormat!=GL_TEXTURE_2D ) {
		u = v = 1.0;
		u1=(float)videoFrame.width-2.0;
		v1=(float)videoFrame.height-2.0;
	}
	else {
		u = pixborderW = 1.0/(float)videoFrame.width;
		v = pixborderH = 1.0/(float)videoFrame.height;
		u1=(float)videoFrame.width/(float)videoFrame.texWidth-pixborderW;
		v1=(float)videoFrame.height/(float)videoFrame.texHeight-pixborderH;
	}

	GLfloat war = (GLfloat)width/(GLfloat)height;

	if ( !hasVideo ) {
		left = -1.0;
		right = 1.0;
		top = -1.0;
		bottom = 1.0;
		videoTopLeft.setX( 0 );
		videoTopLeft.setY( 0 );
 		videoBottomRight.setX( width );
		videoBottomRight.setY( height );
		videoWidth = 2.0;
		videoHeight = 2.0;
	}
	else if ( war<videoFrame.aspectRatio ) {
		left = -1.0;
		right = 1.0;
		top = -war/videoFrame.aspectRatio;
		bottom = war/videoFrame.aspectRatio;
		videoTopLeft.setX( 0 );
		videoTopLeft.setY( (int)((1.0-top)*height/2.0) );
		videoBottomRight.setX( width );
		videoBottomRight.setY( height-videoTopLeft.y() );
		videoWidth = 2.0;
		videoHeight = bottom-top;
	}
	else {
		top = -1.0;
		bottom = 1.0;
		left = -videoFrame.aspectRatio/war;
		right = videoFrame.aspectRatio/war;
		videoTopLeft.setX( (int)((1.0-right)*width/2.0) );
		videoTopLeft.setY( 0 );
		videoBottomRight.setX( width-videoTopLeft.x() );
		videoBottomRight.setY( height );
		videoWidth = right-left;
		videoHeight = 2.0;
	}

	//GLfloat zoomX=m_currentZoomX/100.0, zoomY=m_currentZoomY/100.0;
	//glScalef( zoomX, zoomY, 1.0 );


	if ( videoFrame.format==XINE_VORAW_YV12 ) {
		glEnable( GL_FRAGMENT_PROGRAM_ARB );
		func->mglBindProgram( GL_FRAGMENT_PROGRAM_ARB, fpyv12 );
		func->mglActiveTexture( GL_TEXTURE0 );
		glBindTexture( videoFrame.texFormat, videoTexture[0] );
		func->mglActiveTexture( GL_TEXTURE1 );
		glBindTexture( videoFrame.texFormat, videoTexture[1] );
		func->mglActiveTexture( GL_TEXTURE2 );
		glBindTexture( videoFrame.texFormat, videoTexture[2] );
	}
	else if ( videoFrame.format==XINE_VORAW_YUY2 ) {
		if ( videoFrame.texFormat==GL_TEXTURE_RECTANGLE_ARB ) {
			glEnable( GL_TEXTURE_RECTANGLE_ARB );
			func->mglUseProgramObject(PHandle);
			func->mglActiveTexture( GL_TEXTURE0 );
			glBindTexture( videoFrame.texFormat, videoTexture[0] );
		}
		else {
			glEnable( GL_FRAGMENT_PROGRAM_ARB );
			func->mglBindProgram( GL_FRAGMENT_PROGRAM_ARB, fpyuy2 );
			func->mglActiveTexture( GL_TEXTURE0 );
			glBindTexture( videoFrame.texFormat, videoTexture[0] );
			func->mglActiveTexture( GL_TEXTURE1 );
			glBindTexture( videoFrame.texFormat, videoTexture[1] );
		}
	}
	else {
		func->mglActiveTexture( GL_TEXTURE0 );
		glBindTexture( videoFrame.texFormat, videoTexture[0] );
	}

	glColor4f( 1,1,1,1);
	glBegin (GL_QUADS);
		glTexCoord2d( u, v );
		glVertex3f( left, top, -5.);
		glTexCoord2d( u, v1 );
		glVertex3f( left, bottom, -5.);
		glTexCoord2d( u1, v1 );
		glVertex3f( right, bottom, -5.);
		glTexCoord2d( u1, v );
		glVertex3f( right, top, -5.);
	glEnd ();

	if ( videoFrame.format==XINE_VORAW_YV12 || videoFrame.format==XINE_VORAW_YUY2 ) {
		if ( videoFrame.texFormat==GL_TEXTURE_RECTANGLE_ARB ) {
			func->mglUseProgramObject( 0 );
			glBindTexture( GL_TEXTURE_RECTANGLE_ARB, 0 );
			glDisable( GL_TEXTURE_RECTANGLE_ARB );
		}
		else {
			func->mglBindProgram( GL_FRAGMENT_PROGRAM_ARB, 0 );
			glDisable( GL_FRAGMENT_PROGRAM_ARB );
		}
	}

	func->mglActiveTexture(GL_TEXTURE1);
	glBindTexture( GL_TEXTURE_2D, 0 );
	func->mglActiveTexture(GL_TEXTURE2);
	glBindTexture( GL_TEXTURE_2D, 0 );
	if ( drawOvl ) {
		glPushMatrix();
		glTranslatef( -videoWidth/2.0, -videoHeight/2.0, 0.0 );
		GLfloat ox, oy, ow, oh;
		GLfloat pixH=videoHeight/(float)videoFrame.height, pixW=videoWidth/(float)videoFrame.width;
		int i;
		for ( i=0; i<videoFrame.numOverlays; ++i ) {
			pixborderW = 1.0/videoFrame.ovlRect[i].width();
			pixborderH = 1.0/videoFrame.ovlRect[i].height();
			u=pixborderW; v=pixborderH;
			u1=videoFrame.ovlRect[i].width()/(float)videoFrame.ovlTexSize[i].width()-pixborderW;
			v1=videoFrame.ovlRect[i].height()/(float)videoFrame.ovlTexSize[i].height()-pixborderH;
			ox = videoFrame.ovlRect[i].x()*pixW;
			oy = videoFrame.ovlRect[i].y()*pixH;
			ow = videoFrame.ovlRect[i].width()*pixW;
			oh = videoFrame.ovlRect[i].height()*pixH;
			func->drawQuadTexture( ovlTexture[i], ox, oy, ow, oh, u, v, u1, v1 );
		}
		glPopMatrix();
	}

	glPopMatrix();

	func->mglActiveTexture(GL_TEXTURE0);
	glBindTexture( GL_TEXTURE_2D, 0 );
}



#define BUFFER_OFFSET(i) ((char *)NULL + (i))
void GLVideo::updateVideoTexture( bool resetTex, unsigned char *frameBuf, const int &textureUnit,
	const GLuint &textureID, const GLenum &format, const int &frameW, const int &frameH,
	const int &texW, const int &texH, const int &nBytes, const GLint &filter )
{
	if ( resetTex && videoFrame.texFormat==GL_TEXTURE_RECTANGLE_ARB ) {
		int i=func->mglGetUniformLocation( PHandle, "YUVtex" );
		func->mglUniform1i( i, 0 );  // Bind YUVtex to texture unit 0
	}
	func->mglActiveTexture( GL_TEXTURE0+textureUnit );
	glBindTexture( videoFrame.texFormat, textureID );
	if ( resetTex ) {
		glTexParameteri( videoFrame.texFormat, GL_TEXTURE_MIN_FILTER, filter );
		glTexParameteri( videoFrame.texFormat, GL_TEXTURE_MAG_FILTER, filter );
		glTexImage2D( videoFrame.texFormat, 0, nBytes, texW, texH, 0, format, GL_UNSIGNED_BYTE, NULL );
	}
	if ( func->doPBO() ) {
		func->mglBindBuffer( GL_PIXEL_UNPACK_BUFFER_ARB, videoPBO );
		func->mglBufferData( GL_PIXEL_UNPACK_BUFFER_ARB, frameW*frameH*nBytes, NULL, GL_STREAM_DRAW );
		void *mem = func->mglMapBuffer( GL_PIXEL_UNPACK_BUFFER_ARB, GL_WRITE_ONLY );
		memcpy( mem, frameBuf, frameW*frameH*nBytes );
		func->mglUnmapBuffer( GL_PIXEL_UNPACK_BUFFER_ARB );
		glTexSubImage2D( videoFrame.texFormat, 0, 0, 0, frameW, frameH, format, GL_UNSIGNED_BYTE, BUFFER_OFFSET(0) );
	}
	else
		glTexSubImage2D( videoFrame.texFormat, 0, 0, 0, frameW, frameH, format, GL_UNSIGNED_BYTE, frameBuf );
	if ( func->doPBO() )
		func->mglBindBuffer( GL_PIXEL_UNPACK_BUFFER_ARB, 0 );
}



void GLVideo::updateFrame( const int &frame_format, const int &frame_width,
		const int &frame_height, const double &frame_aspect, void *data0, void *data1, void *data2 )
{
	bool ret = true;
	bool reset = false;
	int tex_w=frame_width, tex_h=frame_height;

	if ( tex_w<2 || tex_h<2 ) {
		return;
	}

	vw->makeCurrent();

	if ( videoFrame.width!=tex_w || videoFrame.height!=tex_h || videoFrame.format!=frame_format ) {
		reset = true;
		if ( frame_format==XINE_VORAW_YUY2 && PHandle ) {
			videoFrame.texFormat = GL_TEXTURE_RECTANGLE_ARB;
		}
		else {
			videoFrame.texFormat = GL_TEXTURE_2D;
			if ( !func->doTextureRect() ) {
				tex_w = 2; tex_h = 2;
				while ( tex_w<frame_width ) tex_w <<= 1;
				while ( tex_h<frame_height ) tex_h <<= 1;
			}
		}
	}

	switch ( frame_format ) {
		case XINE_VORAW_YV12: {
			if ( !data0 || !data1 || !data2 ) {
				ret = false;
				break;
			}
			if ( reset ) kDebug() << QString("YV12 - framesize: %1x%2, texsize: %3x%4").arg(frame_width).arg(frame_height).arg(tex_w).arg(tex_h);
			updateVideoTexture( reset, (unsigned char*)data0, 0, videoTexture[0], GL_LUMINANCE, frame_width, frame_height, tex_w, tex_h, 1, GL_LINEAR );
			updateVideoTexture( reset, (unsigned char*)data1, 1, videoTexture[1], GL_LUMINANCE, frame_width/2, frame_height/2, tex_w/2, tex_h/2, 1, GL_LINEAR );
			updateVideoTexture( reset, (unsigned char*)data2, 2, videoTexture[2], GL_LUMINANCE, frame_width/2, frame_height/2, tex_w/2, tex_h/2, 1, GL_LINEAR );
			break;
		}
		case XINE_VORAW_YUY2: {
			if ( !data0 ) {
				ret = false;
				break;
			}
			if ( PHandle ) {
				if ( reset ) kDebug() << QString("YUY2 shader - framesize: %1x%2, texsize: %3x%4").arg(frame_width).arg(frame_height).arg(tex_w).arg(tex_h);
				updateVideoTexture( reset, (unsigned char*)data0, 0, videoTexture[0], GL_LUMINANCE_ALPHA, frame_width, frame_height, tex_w, tex_h, 2, GL_NEAREST );
			}
			else {
				if ( reset ) kDebug() << QString("YUY2 - framesize: %1x%2, texsize: %3x%4").arg(frame_width).arg(frame_height).arg(tex_w).arg(tex_h);
				updateVideoTexture( reset, (unsigned char*)data0, 0, videoTexture[0], GL_LUMINANCE_ALPHA, frame_width, frame_height, tex_w, tex_h, 2, GL_LINEAR );
				updateVideoTexture( reset, (unsigned char*)data0, 1, videoTexture[1], GL_RGBA, frame_width/2, frame_height, tex_w/2, tex_h, 4, GL_LINEAR );
			}
			break;
		}
		case XINE_VORAW_RGB: {
			if ( !data0 ) {
				ret = false;
				break;
			}
			if ( reset ) kDebug() << QString("RGB - framesize: %1x%2, texsize: %3x%4").arg(frame_width).arg(frame_height).arg(tex_w).arg(tex_h);
			updateVideoTexture( reset, (unsigned char*)data0, 0, videoTexture[0], GL_RGB, frame_width, frame_height, tex_w, tex_h, 3, GL_LINEAR );
			break;
		}
		default: {
			ret = false;
			break;
		}
	}
	if ( ret ) {
		videoFrame.width = frame_width;
		videoFrame.height = frame_height;
		videoFrame.format = frame_format;
		videoFrame.aspectRatio = frame_aspect;
		if ( reset ) {
			videoFrame.texWidth = tex_w;
			videoFrame.texHeight = tex_h;
		}
		emit wantRepaint();
	}
}



void GLVideo::updateOverlays( const int &num_ovl, raw_overlay_t *overlay_array )
{
	int i, tw, th;
	raw_overlay_t *ovl;

	vw->makeCurrent();

	for ( i=0; i<num_ovl; ++i ) {
		ovl = &overlay_array[i];
		func->textureFromData( ovlTexture[i], ovl->ovl_w, ovl->ovl_h, ovl->ovl_rgba, tw, th );
		videoFrame.ovlTexSize[i].setWidth( tw );
		videoFrame.ovlTexSize[i].setHeight( th );
		videoFrame.ovlRect[i] = QRect( ovl->ovl_x, ovl->ovl_y, ovl->ovl_w, ovl->ovl_h );
	}
	videoFrame.numOverlays = num_ovl;
	emit wantRepaint();
}



